from dataclasses import dataclass
from typing import Dict, Union

import affine
from osgeo import gdal
from osgeo import osr
import rasterio
import numpy as np

from rasterio.features import shapes
from rasterio.io import BufferedDatasetWriter, DatasetWriter
from shapely.geometry import shape
from stitch_n_split.geo_info import get_affine_transform, get_pixel_resolution
from stitch_n_split.split.mesh import ImageNonOverLapMesh, ImageOverLapMesh


@dataclass
class Bitmap:
    array: np.ndarray
    transform: affine.Affine


def create_bitmap(
    mesh: Union[ImageNonOverLapMesh, ImageOverLapMesh], geometry_collection: list
):
    """

    :param mesh:
    :param geometry_collection:
    :return:
    """
    intermediate_bitmap = np.empty((0,))
    intermediate_bitmap = np.concatenate(
        (intermediate_bitmap, geometry_collection), axis=0
    )

    for grid in mesh.extent():
        transform = get_affine_transform(
            grid["extent"][0],
            grid["extent"][-1],
            *get_pixel_resolution(mesh.mesh_transform),
        )
        bitmap_array = rasterio.features.rasterize(
            ((g, 255) for g in intermediate_bitmap),
            out_shape=mesh.grid_size,
            transform=transform,
        )
        yield Bitmap(bitmap_array, transform)


def image_to_collection_generator(
    image: np.ndarray, transform: affine.Affine, is_shape=False, **kwargs
) -> Dict:
    """

    :param is_shape:
    :param transform:
    :param image:
    :return:
    """
    for i, (s, v) in enumerate(
        shapes(
            image.astype(rasterio.uint8),
            mask=(image > 0),
            connectivity=8,
            transform=transform,
        )
    ):
        yield {
            "geometry": shape(s) if is_shape else s,
            "properties": {"id": v, **kwargs},
        }


def get_geo_referenced_image_obj(
    copy_from: Union[BufferedDatasetWriter, DatasetWriter], copy_to: np.ndarray
):
    """

    :param copy_from:
    :param copy_to:
    :return:
    """
    bands = copy_to.ndim if copy_to.ndim > 2 else 1
    geo_referenced_image = rasterio.open(
        "",
        mode="w",
        driver=copy_from.driver,
        width=copy_from.width,
        height=copy_from.height,
        crs=copy_from.crs,
        transform=copy_from.transform,
        dtype=copy_to.dtype,
        count=bands,
    )
    return geo_referenced_image


def copy_geo_reference_to_image(
    copy_from: Union[BufferedDatasetWriter, DatasetWriter],
    copy_to: np.ndarray,
    save_to: str,
):
    """

    :param copy_from:
    :param copy_to:
    :param save_to:
    :return:
    """
    bands = copy_to.ndim if copy_to.ndim > 2 else 1
    geo_referenced_image = rasterio.open(
        save_to,
        mode="w",
        driver=copy_from.driver,
        width=copy_from.width,
        height=copy_from.height,
        crs=copy_from.crs,
        transform=copy_from.transform,
        dtype=copy_to.dtype,
        count=bands,
    )
    if bands > 2:
        for band in range(copy_to.shape[2]):
            geo_referenced_image.write(copy_to[:, :, band], band + 1)
    else:
        geo_referenced_image.write(copy_to, 1)

    copy_from.close()
    geo_referenced_image.close()


def save_multi_band(
    image: np.ndarray,
    geo_transform: affine.Affine,
    gdal_unit,
    epsg: int,
    output_file_name: str,
):
    """

    :param image: image must be of the format h X w X bands
    :param geo_transform:
    :param gdal_unit:
    :param epsg:
    :param output_file_name:
    :return:
    """
    assert image.ndim == 3, f"Expected to have 3 dim got {image.ndim}"

    x, y, z = image.shape
    dst_ds = gdal.GetDriverByName("GTiff").Create(output_file_name, y, x, z, gdal_unit)

    dst_ds.SetGeoTransform(geo_transform)
    srs = osr.SpatialReference()
    srs.ImportFromEPSG(epsg)
    dst_ds.SetProjection(srs.ExportToWkt())

    for idx in range(0, z):
        dst_ds.GetRasterBand(idx + 1).WriteArray(image[:, :, idx])

    dst_ds.FlushCache()



from typing import Union

import numpy as np

from py_gis_utility.ops.math_ops import (
    new_perpendicular_point_to_line_segment,
    new_coordinate_based_on_angle_and_distance,
    new_point_after_certain_distance,
    euclidean,
    euclidean_between_two_sets,
)
from py_gis_utility.helper import is_line_segment_3d, is_point_3d


def get_new_perpendicular_point_with_common_distance_all_to_line_segment(
    line_segments: np.ndarray,
    distance_from_the_line: Union[int, float] = 10,
    return_pandas: bool = True,
):
    """
    :param return_pandas:
    :param line_segments: array of shape [number_of_line_segments, 2, 2]
    :param distance_from_the_line: how far the new point to create from the reference
    :return:

    """

    assert type(line_segments) is np.ndarray and type(distance_from_the_line) in [
        float,
        int,
    ], (
        f"Expected to have input type ['np.ndarray', '[int, float]'] "
        f"got {type(line_segments), type(distance_from_the_line)}"
    )

    assert is_line_segment_3d(line_segments), (
        f"Expected line segments to be either '[n_values, 2, 2]' for n_dim == 3 "
        f"got {line_segments.shape} for n_dim == {line_segments.ndim}"
    )

    assert (
        type(distance_from_the_line) in [float, int] and distance_from_the_line >= 0.0
    ), (
        "Expected distance_from_the_line to be of type 'float or int' and 'non zero'"
        f"got {type(distance_from_the_line), distance_from_the_line}"
    )

    common_distance_from_the_line = (
        np.ones((line_segments.shape[0], 1, 1)) * distance_from_the_line
    )

    return new_perpendicular_point_to_line_segment(
        line_segments, common_distance_from_the_line
    )


def get_new_perpendicular_point_with_custom_distance_to_every_line_segment(
    line_segments: np.ndarray, distance_from_the_line: np.ndarray
):
    """
    :param line_segments: array of shape [number_of_line_segments, 2, 2]
    :param distance_from_the_line: how far the new point to create from the reference
    :return:

    """
    return new_perpendicular_point_to_line_segment(
        line_segments, distance_from_the_line
    )


def get_new_coordinates_with_custom_angle_and_distance_for_every_point(
    points: np.ndarray, angle_in_degree: np.ndarray, distance: np.ndarray
) -> np.ndarray:
    """
    :param points: array of shape [number_of_line_segments, 1, 2]
    :param angle_in_degree: array of shape [number_of_points, 1, 1]
    :param distance: array of shape [number_of_points, 1, 1]
    :return:
    """
    return new_coordinate_based_on_angle_and_distance(points, angle_in_degree, distance)


def get_new_coordinates_with_common_angle_and_distance_to_all_points(
    points: np.ndarray, angle_in_degree: Union[float, int], distance: Union[float, int]
) -> np.ndarray:
    """
    :param points: array of shape [number_of_line_segments, 1, 2]
    :param angle_in_degree:
    :param distance:
    :return:
    """

    assert (
        type(points) is np.ndarray
        and type(distance) in [float, int]
        and type(angle_in_degree) in [float, int]
    ), (
        "Expected to have input type ['np.ndarray', '[float, int]', '[float, int]']"
        f"got {type(points), type(distance), type(angle_in_degree)}"
    )

    assert is_point_3d(points), (
        f"Expected points to be either '[n_points, 1, 2]' for n_dim == 3"
        f"got {points.shape} for n_dim == {points.ndim}"
    )

    return new_coordinate_based_on_angle_and_distance(
        points,
        (np.ones((points.shape[0], 1, 1)) * angle_in_degree),
        (np.ones((points.shape[0], 1, 1)) * distance),
    )


def get_points_with_custom_distance_for_every_line_segments(
    line_segments: np.ndarray, distance_from_start: np.ndarray
) -> np.ndarray:
    """

    :param line_segments: array of shape [number_of_line_segments, 2, 2]
    :param distance_from_start: array specifying the distance to compute the point
            shape [number_of_line_segments, 1, 1]
    :return:
    """
    return new_point_after_certain_distance(line_segments, distance_from_start)


def get_new_points_with_same_distance_for_all_line_segments(
    line_segments: np.ndarray, distance_from_start: Union[float, int]
) -> np.ndarray:
    """

    :param line_segments: array of shape [number_of_line_segments, 2, 2]
    :param distance_from_start:
    :return:
    """
    assert type(line_segments) is np.ndarray and type(distance_from_start) in [
        float,
        int,
    ], f"Expected to have input type ['np.ndarray', '[int, float]'] got {type(line_segments), type(distance_from_start)}"

    assert is_line_segment_3d(line_segments), (
        f"Expected line segments to be either '[n_values, 2, 2]' for n_dim == 3"
        f"got {line_segments.shape} for n_dim == {line_segments.ndim}"
    )

    assert type(distance_from_start) in [float, int] and distance_from_start >= 0.0, (
        "Expected distance_from_the_line to be of type 'float or int' and 'non zero'"
        f"got {type(distance_from_start), distance_from_start}"
    )

    common_distance_from_start = (
        np.ones((line_segments.shape[0], 1, 1)) * distance_from_start
    )

    return new_point_after_certain_distance(line_segments, common_distance_from_start)


def get_euclidean(coordinates_a: np.ndarray) -> np.ndarray:
    """

    :param coordinates_a:
    :return:
    """
    return euclidean(coordinates_a)


def get_euclidean_within_two(
    coordinates_a: np.ndarray, coordinates_b: np.ndarray
) -> np.ndarray:
    """

    :param coordinates_a:
    :param coordinates_b:
    :return:
    """
    return euclidean_between_two_sets(coordinates_a, coordinates_b)



import numpy as np
from scipy.spatial.distance import pdist, squareform

from py_gis_utility.helper import (
    is_line_segment_3d,
    is_value_3d,
    is_point_3d,
)


def angle_between_vector(v1: tuple, v2: tuple):
    """
    two vectors have either the same direction -  https://stackoverflow.com/a/13849249/71522
    :param v1:
    :param v2:
    :return:
    """
    v1_u = unit_vector(v1)
    v2_u = unit_vector(v2)
    return np.arccos(np.clip(np.dot(v1_u, v2_u), -1.0, 1.0))


def vector(p1: np.ndarray, p2: np.ndarray) -> np.ndarray:
    """

    :param p1:
    :param p2:
    :return:
    """
    return np.subtract(p1, p2)


def magnitude(vec: np.ndarray, axis=-1) -> np.ndarray:
    return np.linalg.norm(vec, axis=axis)


def unit_vector(vec: np.ndarray, axis=-1) -> np.ndarray:
    """

    :param axis:
    :param vec:
    :return: unit vector
    """
    return np.divide(
        vec, np.concatenate([magnitude(vec, axis)[:, :, np.newaxis]] * 2, axis=axis)
    )


def euclidean_between_two_sets(
    coordinates_a: np.ndarray, coordinates_b: np.ndarray
) -> np.ndarray:
    """
    This function will return set of distances from coordinates_b to coordinates_a
    ex -
        coordinates_b = [[0, 0], [1, 1]]
        coordinates_a = [[2, 1], [3, 1], [2, 4], [5, 2]]

        return
            sets of distances from [0, 0] to all the coordinates present in coordinates_a
            sets of distances from [1, 1] to all the coordinates present in coordinates_a

    visually -
        | distance from [0, 0] to [2, 1]   | distance from [1, 1] to [2, 1] |
        | distance from [0, 0] to [3, 1]   | distance from [1, 1] to [3, 1] |
        | distance from [0, 0] to [2, 4]   | distance from [1, 1] to [2, 4] |
        | distance from [0, 0] to [5, 2]   | distance from [1, 1] to [5, 2] |

    resulting in shape of [n_coordinates_a, n_coordinates_b]

    :param coordinates_a: array of shape [n_coordinates_a, 2]
    :param coordinates_b: array of shape [n_coordinates_b, 2]
    :return: array of shape [n_coordinates_a, n_coordinates_b]
    """

    assert (
        type(coordinates_a) is np.ndarray and type(coordinates_b) is np.ndarray
    ), f"Expected to have input type 'np.ndarray' got {type(coordinates_a)}, {type(coordinates_b)}"

    assert (
        coordinates_b.shape[-1] == 2 and coordinates_b.ndim == 2
    ), f"Expected input_coordinates to have shape '[number of points, 2]' got {coordinates_b.shape}"

    assert (
        coordinates_a.shape[-1] == 2 and coordinates_a.ndim == 2
    ), f"Expected input_coordinates to have shape '[number of points, 2]' got {coordinates_a.shape}"

    distances = np.linalg.norm(
        np.concatenate(
            [coordinates_a[np.newaxis, :, :]] * coordinates_b.shape[0], axis=0
        )
        - coordinates_b[:, np.newaxis, :],
        axis=-1,
    ).T
    return distances


def euclidean(coordinates_a: np.ndarray) -> np.ndarray:
    """
    This function will return set of distances from coordinates_a to coordinates_a
    ex -
        coordinates_a = [[0, 0], [1, 1]]

        return
            sets of distances from [0, 0] to all the coordinates present in coordinates_a
            sets of distances from [1, 1] to all the coordinates present in coordinates_a

    visually -
        | distance from [0, 0] to [0, 0]   | distance from [1, 1] to [0, 0] |
        | distance from [0, 0] to [1, 1]   | distance from [1, 1] to [1, 1] |

    resulting in shape of [n_coordinates_a, n_coordinates_a]

    :param coordinates_a: array of shape [n_coordinates_a, 2]
    :return: array of shape [n_coordinates_a, n_coordinates_a]
    """
    assert type(coordinates_a) is np.ndarray, (
        "Expected to have input type 'np.ndarray'" "got %s",
        (type(coordinates_a),),
    )
    assert (
        type(coordinates_a) is np.ndarray
    ), f"Expected to have input type 'np.ndarray' got {type(coordinates_a)}"

    assert (
        coordinates_a.shape[-1] == 2 and coordinates_a.ndim == 2
    ), f"Expected input_coordinates to have shape '[number of points, 2]' got {coordinates_a.shape}"

    return squareform(pdist(coordinates_a))


def perpendicular_distance_from_point_to_line_segment_in_2d(
    line_segment: np.ndarray, coordinates: np.ndarray
) -> np.ndarray:
    """
    The function will compute perpendicular distance for the coordinates value present in coordinates to the given
    input line segment

    :param line_segment: array of shape [number_of_line_segments, 2, 2]

            If the  line segment to which perpendicular distances are to be computed then pass it as
            follows :
                -- the dimension [number_of_line_segments, 2, 2] are [
                                                                        [
                                                                        [start_line_segment_1_x, start_line_segment_1_y],
                                                                        [end_line_segment_1_x,   end_line_segment_1_y]
                                                                        ],
                                                                        [
                                                                        [start_line_segment_2_x, start_line_segment_2_y],
                                                                        [end_line_segment_2_x,   end_line_segment_2_y]
                                                                        ],
                                                                        ....
                                                                        ...
                                                                        [
                                                                        [start_line_segment_n_x, start_line_segment_n_y],
                                                                        [end_line_segment_n_x,   end_line_segment_n_y]
                                                                        ],
                                                                    ]
    :param coordinates: array of shape [number of points, 2]
    :return: array of shape [number of segments, number of points]

    if single line segment is passed for computation, i.e. input with dim [1, 2, 2] then expect the output in
        format:
            | distance from line_Segment_1 to coordinate[0]  .... | .... distance from line_Segment_1 to coordinate[n] |

        if multiple line segments are passed for computation, i.e. input with dim [n_segments, 2, 2] then expect
         the output in format:
            | distance from line_Segment_1 to coordinate[0]  ....  | ... distance from line_Segment_1 to coordinate[n] |
            | distance from line_Segment_2 to coordinate[0]  ....  | ... distance from line_Segment_2 to coordinate[n] |
            ....
            ...
            | distance from line_Segment_n to coordinate[0]  ....  | ...distance from line_Segment_n to coordinate[n] |

    """

    # https://stackoverflow.com/a/53176074/7984359
    # https://math.stackexchange.com/questions/1300484/distance-between-line-and-a-point
    # https://www.geeksforgeeks.org/minimum-distance-from-a-point-to-the-line-segment-using-vectors/
    # https://en.wikipedia.org/wiki/Distance_from_a_point_to_a_line#Line_defined_by_two_points

    assert (
        type(line_segment) is np.ndarray and type(coordinates) is np.ndarray
    ), f"Expected to have input type 'np.ndarray' got {type(line_segment)} and {type(coordinates)}"

    assert coordinates.ndim == 2 and coordinates.shape[-1] == 2, (
        f"Expected coordinates to be '2' dimensional and have shape '[number of point, 2]' got {coordinates.shape},"
        f" {coordinates.ndim}"
    )

    assert is_line_segment_3d(line_segment), (
        f"Expected line segments to be either '[n_line_segments, 2, 2]' for n_dim == 3"
        f"got {line_segment.shape} for n_dim == {line_segment.ndim}"
    )

    coordinates_repeat_copy = np.concatenate(
        [coordinates[np.newaxis, :, :]] * line_segment.shape[0], axis=0
    )

    dp = line_segment[:, 1:2, :] - line_segment[:, 0:1, :]
    st = dp[:, :, 0:1] ** 2 + dp[:, :, 1:2] ** 2

    u = (
        (coordinates_repeat_copy[:, :, 0:1] - line_segment[:, 0:1, 0:1]) * dp[:, :, 0:1]
        + (coordinates_repeat_copy[:, :, 1:2] - line_segment[:, 0:1, 1:2])
        * dp[:, :, 1:2]
    ) / st

    u[u > 1.0] = 1.0
    u[u < 0.0] = 0.0

    dx = (line_segment[:, 0:1, 0:1] + u * dp[:, :, 0:1]) - coordinates_repeat_copy[
        :, :, 0:1
    ]
    dy = (line_segment[:, 0:1, 1:2] + u * dp[:, :, 1:2]) - coordinates_repeat_copy[
        :, :, 1:2
    ]

    return np.squeeze(np.sqrt(dx ** 2 + dy ** 2), axis=-1)


def new_perpendicular_point_to_line_segment(
    line_segment: np.ndarray, distance_from_the_line: np.ndarray
):
    """
    Get perpendicular point with reference to start and end point of the segment

    :param line_segment: array of shape [number_of_line_segments, 2, 2]

            If the line segment to which perpendicular distances are to be computed then pass it as
            follows :
                -- the dimension [number_of_line_segments, 2, 2] are [
                                                                        [
                                                                        [start_line_segment_1_x, start_line_segment_1_y],
                                                                        [end_line_segment_1_x,   end_line_segment_1_y]
                                                                        ],
                                                                        [
                                                                        [start_line_segment_2_x, start_line_segment_2_y],
                                                                        [end_line_segment_2_x,   end_line_segment_2_y]
                                                                        ],
                                                                        ....
                                                                        ...
                                                                        [
                                                                        [start_line_segment_n_x, start_line_segment_n_y],
                                                                        [end_line_segment_n_x,   end_line_segment_n_y]
                                                                        ],
                                                                    ]
    :param distance_from_the_line: how far the new point to create from the reference
    :return:(perpendicular points with reference to start, perpendicular points with reference to end)
            -
                return is of shape [number_of_segments, 2, 2]

                    [A_n]                          [C_n]
                    |     line_segment_with       |
                    |-----------------------------|
                    |     index value 'n'         |
                    [B_n]                         [D_n]

                to get points -

                    A_n - perpendicular_with_start[segment_index_value_n, 0, :]
                    B_n - perpendicular_with_start[segment_index_value_n, 1, :]
                    C_n - perpendicular_with_end[segment_index_value_n, 0, :]
                    D_n - perpendicular_with_end[segment_index_value_n, 1, :]

    """
    assert (
        type(line_segment) is np.ndarray and type(distance_from_the_line) is np.ndarray
    ), (
        f"Expected to have input type ['np.ndarray', 'np.ndarray'] got {type(line_segment)}"
        f" and {type(distance_from_the_line)}"
    )

    assert is_line_segment_3d(line_segment), (
        f"Expected line segments to be either '[n_line_segments, 2, 2]' for n_dim == 3 "
        f"got {line_segment.shape} for n_dim == {line_segment.ndim}"
    )

    assert is_value_3d(distance_from_the_line), (
        f"Expected values to be either '[n_values, 1, 1]' for n_dim == 3 or '[1, 1]' for n_dim == 2, "
        f"got {distance_from_the_line.shape} for n_dim == {distance_from_the_line.ndim}"
    )

    assert line_segment.shape[0] == distance_from_the_line.shape[0], (
        f"Expected number of distance to "
        f"be equal to number of line segments,"
        f" got {line_segment.shape[0],distance_from_the_line.shape[0]}"
    )

    assert np.all(
        distance_from_the_line >= 0
    ), f"Expected distance to be 'non zero' got {distance_from_the_line}"

    x1, y1 = line_segment[:, 0:1, 0:1], line_segment[:, 0:1, 1:2]
    x2, y2 = line_segment[:, 1:2, 0:1], line_segment[:, 1:2, 1:2]

    dx = x1 - x2
    dy = y1 - y2

    dist = np.linalg.norm(np.array([dx, dy]), axis=0)

    x_perpendicular = distance_from_the_line * (dx / dist)
    y_perpendicular = distance_from_the_line * (dy / dist)

    point_with_start_as_reference_x = [
        np.array(x1 + y_perpendicular).squeeze(axis=-1),
        np.array(y1 - x_perpendicular).squeeze(axis=-1),
    ]

    point_with_start_as_reference_y = [
        np.array(x1 - y_perpendicular).squeeze(axis=-1),
        np.array(y1 + x_perpendicular).squeeze(axis=-1),
    ]

    point_with_start_as_reference = np.dstack(
        (
            np.array(point_with_start_as_reference_x).squeeze(axis=-1),
            np.array(point_with_start_as_reference_y).squeeze(axis=-1),
        )
    )
    point_with_end_as_reference_x = [
        np.array(x2 - y_perpendicular).squeeze(axis=-1),
        np.array(y2 + x_perpendicular).squeeze(axis=-1),
    ]

    point_with_end_as_reference_y = [
        np.array(x2 + y_perpendicular).squeeze(axis=-1),
        np.array(y2 - x_perpendicular).squeeze(axis=-1),
    ]

    point_with_end_as_reference = np.dstack(
        (
            np.array(point_with_end_as_reference_x).squeeze(axis=-1),
            np.array(point_with_end_as_reference_y).squeeze(axis=-1),
        )
    )

    perpendicular_with_start = point_with_start_as_reference.swapaxes(-1, 1).T
    perpendicular_with_end = point_with_end_as_reference.swapaxes(-1, 1).T

    return perpendicular_with_start, perpendicular_with_end


def new_coordinate_based_on_angle_and_distance(
    points: np.ndarray, angle: np.ndarray, distance: np.ndarray
) -> np.ndarray:
    """
    Given a Point find a new point at an given 'angle' with given 'distance'

          / B [New Point]
         /
        /  angle CAB and distance AB [GIVEN]
       A ------------ C

    # https://math.stackexchange.com/questions/39390/determining-end-coordinates-of-line-with-the-specified-length-and-angle

    :param points: array of shape [number_of_line_segments, 1, 2]
            If the line segment to which perpendicular distances are to be computed then pass it as
            follows :
                -- the dimension [number_of_points, 1, 2] are [
                                                                        [
                                                                        [point_1_x, point_1_y],
                                                                        ],
                                                                        [
                                                                        [point_2_x, point_2_y],
                                                                        ],
                                                                        ....
                                                                        ...
                                                                        [
                                                                        [point_n_x, point_n_y],
                                                                        ],
                                                                    ]
    :param angle: array of shape [number_of_points, 1, 1]
    :param distance: array of shape [number_of_points, 1, 1]
    :return:
    """
    assert (
        type(points) is np.ndarray
        and type(distance) is np.ndarray
        and type(angle) is np.ndarray
    ), (
        "Expected to have input type ['np.ndarray', 'np.ndarray', 'np.ndarray']"
        f"got {type(points), type(distance), type(angle)}"
    )

    assert is_point_3d(points), (
        f"Expected points to be either '[n_points, 1, 2]' for n_dim == 3 "
        f"got {points.shape} for n_dim == {points.ndim}"
    )
    assert is_value_3d(angle), (
        f"Expected values to be either '[n_values, 1, 1]' for n_dim == 3 "
        f"got {angle.shape} for n_dim == {angle.ndim}"
    )
    assert is_value_3d(distance), (
        f"Expected values to be either '[n_values, 1, 1]' "
        f"got {distance.shape} for n_dim == {distance.ndim}"
    )
    assert points.shape[0] == distance.shape[0] == angle.shape[0], (
        f"Expected number of points, number of distance, "
        f"number of angles to have the same count"
        f" got {points.shape[0],distance.shape[0], angle.shape}"
    )

    assert np.all(distance >= 0), f"Expected distance to be 'non zero' got {distance}"

    return np.concatenate(
        [
            points[:, :, 0:1] + (distance * np.cos(angle)),
            points[:, :, 1:2] + (distance * np.sin(angle)),
        ],
        axis=-1,
    )


def new_point_after_certain_distance(
    line_segments: np.ndarray, distance_from_start: np.ndarray
) -> np.ndarray:
    """
    https://math.stackexchange.com/questions/175896/finding-a-point-along-a-line-a-certain-distance-away-from-another-point
    https://math.stackexchange.com/a/426810

    :param distance_from_start: array specifying the distance to compute the point
            shape [number_of_line_segments, 1, 1] or [1, 1]
    :param line_segments: array of shape [number_of_line_segments, 2, 2]

            If the line segment to which perpendicular distances are to be computed then pass it as
            follows :
                -- the dimension [number_of_line_segments, 2, 2] are [
                                                                        [
                                                                        [start_line_segment_1_x, start_line_segment_1_y],
                                                                        [end_line_segment_1_x,   end_line_segment_1_y]
                                                                        ],
                                                                        [
                                                                        [start_line_segment_2_x, start_line_segment_2_y],
                                                                        [end_line_segment_2_x,   end_line_segment_2_y]
                                                                        ],
                                                                        ....
                                                                        ...
                                                                        [
                                                                        [start_line_segment_n_x, start_line_segment_n_y],
                                                                        [end_line_segment_n_x,   end_line_segment_n_y]
                                                                        ],
                                                                    ]
    :return: points on the line segment with distance 'd'

    """
    assert (
        type(line_segments) is np.ndarray and type(distance_from_start) is np.ndarray
    ), f"Expected to have input type ['np.ndarray', 'np.ndarray'] got {type(line_segments), type(distance_from_start)}"

    assert is_line_segment_3d(line_segments), (
        f"Expected line segments to be either '[n_values, 2, 2]' for n_dim == 3 "
        f"got {line_segments.shape} for n_dim == {line_segments.ndim}"
    )

    assert np.all(
        distance_from_start >= 0
    ), f"Expected distance_from_the_line to be  'non negative' got {distance_from_start}"

    assert is_value_3d(distance_from_start), (
        f"Expected values to be either '[n_values, 1, 1]' for n_dim == 3"
        f"got {distance_from_start.shape} for n_dim == {distance_from_start.ndim}"
    )

    assert line_segments.shape[0] == distance_from_start.shape[0], (
        f"Expected input to have same number of count,"
        f" got {line_segments.shape[0],distance_from_start.shape[0]}"
    )

    vec = vector(line_segments[:, 1:2, :], line_segments[:, 0:1, :])
    vec_mag = np.concatenate([magnitude(vec, -1)[:, :, np.newaxis]] * 2, axis=-1)
    return (distance_from_start * vec_mag) * unit_vector(vec)



import os
from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

current_dir = os.path.abspath(os.path.dirname(__file__))
try:
    with open(os.path.join(current_dir, "requirements.txt"), encoding="utf-8") as f:
        install_requires = f.read().split("\n")
except FileNotFoundError:
    install_requires = []


setup(
    name="py-gis-utility",
    version="0.2.2",
    author="Fuzail Palnak",
    author_email="fuzailpalnak@gmail.com",
    url="https://github.com/fuzailpalnak/py-gis-utility",
    description="Utility for image and math related operation in GIS",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    python_requires="~=3.3",
    install_requires=install_requires,
    keywords=["GIS", "Image"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
    ],
)



import affine
import cv2
import numpy as np
from typing import Dict, Union
from osgeo import gdal
from geopandas import GeoDataFrame
from rasterio.io import BufferedDatasetWriter, DatasetWriter

from py_gis_utility.helper import (
    read_image_with_geo_transform,
    extract_geometry_from_data_frame_row,
    read_data_frame,
    create_mesh,
)
from py_gis_utility.ops.image_ops import (
    image_to_collection_generator,
    copy_geo_reference_to_image,
    create_bitmap,
    save_multi_band,
)


def image_obj_to_coordinates_generator(
    image: Union[BufferedDatasetWriter, DatasetWriter], **kwargs
) -> Dict:
    """

    :param image:
    :param kwargs:
    :return:
    """
    return image_to_collection_generator(
        image.read(), image.transform, is_shape=False, crs=image.crs, **kwargs
    )


def image_path_to_coordinates_generator(image_path: str, **kwargs) -> Dict:
    """

    :param image_path:
    :return:
    """
    return image_obj_to_coordinates_generator(
        read_image_with_geo_transform(image_path), **kwargs
    )


def image_obj_to_shape_generator(
    image: Union[BufferedDatasetWriter, DatasetWriter], **kwargs
) -> Dict:
    """

    :param image:
    :return:
    """
    return image_to_collection_generator(
        image.read(), image.transform, is_shape=True, crs=image.crs, **kwargs
    )


def image_path_to_shape_generator(image_path: str, **kwargs) -> Dict:
    """

    :param image_path:
    :return:
    """
    return image_obj_to_shape_generator(
        read_image_with_geo_transform(image_path), **kwargs
    )


def copy_geo_reference(
    copy_from_path: str, copy_to_path: str, save_to: str, is_binary: bool = True
):
    """

    :param is_binary:
    :param save_to:
    :param copy_from_path:
    :param copy_to_path:
    :return:
    """
    destination_img = (
        cv2.imread(copy_to_path, cv2.IMREAD_GRAYSCALE)
        if is_binary
        else cv2.cvtColor(cv2.imread(copy_to_path), cv2.COLOR_BGR2RGB)
    )
    copy_geo_reference_to_image(
        read_image_with_geo_transform(copy_from_path),
        destination_img,
        save_to,
    )


def shape_geometry_to_bitmap_from_data_frame_generator(
    data_frame: GeoDataFrame,
    output_image_size: tuple,
    pixel_resolution: tuple,
    allow_output_to_overlap: bool = True,
):
    """

    :param data_frame:
    :param output_image_size:
    :param pixel_resolution:
    :param allow_output_to_overlap:
    :return:
    """
    bitmap_generator = create_bitmap(
        create_mesh(
            data_frame.total_bounds,
            output_image_size,
            pixel_resolution,
            is_overlap=allow_output_to_overlap,
        ),
        list(extract_geometry_from_data_frame_row(data_frame)),
    )

    return bitmap_generator


def shape_geometry_to_bitmap_from_shape_geometry_path_generator(
    shape_geometry_file_path: str,
    output_image_size: tuple,
    pixel_resolution: tuple,
    allow_output_to_overlap: bool = True,
):
    """

    :param shape_geometry_file_path:
    :param output_image_size:
    :param pixel_resolution:
    :param allow_output_to_overlap:
    :return:
    """
    return shape_geometry_to_bitmap_from_data_frame_generator(
        read_data_frame(shape_geometry_file_path),
        output_image_size,
        pixel_resolution,
        allow_output_to_overlap,
    )


def save_8bit_multi_band(
    image: np.ndarray, geo_transform: affine.Affine, epsg: int, output_file_name: str
):
    """

    :param image: image must be of the format h X w X bands
    :param geo_transform:
    :param epsg:
    :param output_file_name:
    :return:
    """
    save_multi_band(image, geo_transform, gdal.GDT_Byte, epsg, output_file_name)


def save_16bit_multi_band(
    image: np.ndarray, geo_transform: affine.Affine, epsg: int, output_file_name: str
):
    """

    :param image: image must be of the format h X w X bands
    :param geo_transform:
    :param epsg:
    :param output_file_name:
    :return:
    """
    save_multi_band(image, geo_transform, gdal.GDT_UInt16, epsg, output_file_name)



from typing import Union

import affine
import geopandas
import numpy as np
import rasterio
from geopandas import GeoDataFrame
from rasterio.io import BufferedDatasetWriter, DatasetWriter
from shapely.geometry import mapping
from stitch_n_split.geo_info import get_affine_transform
from stitch_n_split.split.mesh import (
    mesh_from_geo_transform,
    ImageNonOverLapMesh,
    ImageOverLapMesh,
)


def concatenate_list_of_numpy_images(images_list: list) -> np.ndarray:
    """

    :param images_list: images must be of the format h X w X bands
    :return:
    """
    return np.concatenate(images_list, axis=-1)


def read_data_frame(path: str) -> GeoDataFrame:
    return geopandas.read_file(path)


def create_mesh(
    mesh_bounds: tuple,
    grid_size: tuple,
    pixel_resolution: tuple,
    is_overlap: bool = False,
) -> Union[ImageNonOverLapMesh, ImageOverLapMesh]:
    """

    :param mesh_bounds:
    :param grid_size:
    :param pixel_resolution:
    :param is_overlap: if set to true will find overlapping grid if any
    :return:
    """
    assert len(mesh_bounds) == 4, (
        f"Expected mesh_bounds to be in format (minx, miny, maxx, maxy) but got "
        f"{mesh_bounds} of size {len(mesh_bounds)}"
    )

    assert len(grid_size) == 2, (
        f"Expected grid_size to be in format (h x w) but got "
        f"{grid_size} of size {len(grid_size)}"
    )

    assert (
        len(pixel_resolution) == 2
    ), f"Expected pixel_resolution to have size 2 but got {len(grid_size)}"

    mesh = mesh_from_geo_transform(
        grid_size=grid_size,
        transform=get_affine_transform(
            mesh_bounds[0], mesh_bounds[-1], *pixel_resolution
        ),
        mesh_bounds=mesh_bounds,
        overlap=is_overlap,
    )
    return mesh


def extract_geometry_from_data_frame_row(row: GeoDataFrame) -> list:
    if "geometry" not in list(row.keys()):
        raise KeyError("Missing Keys, Must have keys ['geometry']")

    feature_geometry = list()

    for geometry in list(mapping(row["geometry"])["features"]):
        feature_geometry.append(geometry["geometry"])

    return feature_geometry


def save_image_with_geo_transform(
    save_path: str, image: np.ndarray, transform: affine.Affine
):
    """

    :param save_path:
    :param image:
    :param transform:
    :return:
    """

    bands = 1 if image.ndim == 2 else image.shape[-1]
    with rasterio.open(
        save_path,
        "w",
        driver="GTiff",
        dtype=rasterio.uint8,
        count=bands,
        width=image.shape[0],
        height=image.shape[1],
        transform=transform,
    ) as dst:
        dst.write(image, indexes=1)


def read_image_with_geo_transform(
    path: str,
) -> Union[BufferedDatasetWriter, DatasetWriter]:
    """

    :param path:
    :return:
    """
    return rasterio.open(path)


def convert_2d_input_to_3d_single_batch_format(input_array: np.ndarray) -> np.ndarray:
    input_array = input_array[np.newaxis, :, :]
    return input_array


def convert_1d_coordinates_to_2d_single_batch_format(
    input_coordinates: np.ndarray,
) -> np.ndarray:
    input_coordinates = input_coordinates[np.newaxis, :]
    return input_coordinates


def minimum_in_matrix(input_matrix: np.ndarray, find_minimum_in_axis=1):
    """

    :param input_matrix:
    :param find_minimum_in_axis:
    :return:
    """
    assert (
        input_matrix.shape[-1] == 2 and input_matrix.ndim == 2
    ), f"Expected input_coordinates to have shape '[number of points, 2]'got {input_matrix.shape}"

    assert find_minimum_in_axis in [
        1,
        2,
    ], f"Expected shortest_distance_axis to be in '[1, 2]' got {find_minimum_in_axis}"
    minimum = np.argmin(input_matrix, axis=find_minimum_in_axis)

    return minimum, input_matrix[minimum]


def extract_index(
    from_input: np.ndarray, value: np.ndarray, along_axis: int = 0
) -> np.ndarray:
    """

    :param along_axis:
    :param from_input:
    :param value:
    :return:
    """

    assert (
        type(from_input) is np.ndarray and type(value) is np.ndarray
    ), f"Expected to have input type 'np.ndarray' got {type(from_input), type(value)}"
    assert from_input.ndim == 3 and (from_input.shape[-2], from_input.shape[-1]) == (
        2,
        2,
    ), f"Expected from_input coordinates to be either '[n_from_input, 2, 2]' got {from_input.ndim, from_input.shape}"

    assert (
        0 <= along_axis <= from_input.ndim
    ), f"Expected along_axis to be in range ['0' and '%s'] got {from_input.ndim, along_axis}"

    if value.ndim == 2 and from_input.ndim == 3:
        value = value[np.newaxis, :, :]

    assert (
        from_input.ndim == value.ndim
    ), f"Expected to have same number of dimensions got {from_input.ndim, value.ndim}"

    return np.unique(np.where(from_input == value)[along_axis])


def get_dimension(input_array: np.ndarray) -> int:
    return input_array.ndim


def get_coordinate_structure(input_array: np.ndarray) -> tuple:
    return input_array.shape[-2], input_array.shape[-1]


def is_input_3d(input_array: np.ndarray) -> bool:
    return True if get_dimension(input_array) == 3 else False


def is_input_2d(input_array: np.ndarray) -> bool:
    return True if get_dimension(input_array) == 2 else False


def is_line_segment(input_array: np.ndarray) -> bool:
    return True if (get_coordinate_structure(input_array) == (2, 2)) else False


def is_value(input_array: np.ndarray) -> bool:
    return True if (get_coordinate_structure(input_array) == (1, 1)) else False


def is_point(input_array: np.ndarray) -> bool:
    return True if (get_coordinate_structure(input_array) == (1, 2)) else False


def is_line_segment_3d(line_segments: np.ndarray) -> bool:
    return (
        True
        if (is_input_3d(line_segments) and is_line_segment(line_segments))
        else False
    )


def is_value_3d(values: np.ndarray) -> bool:
    return True if (is_input_3d(values) and is_value(values)) else False


def is_point_3d(points: np.ndarray) -> bool:
    return True if (is_input_3d(points) and is_point(points)) else False


def is_line_segment_2d(line_segments: np.ndarray) -> bool:
    return (
        True
        if (is_input_2d(line_segments) and is_line_segment(line_segments))
        else False
    )


def is_value_2d(values: np.ndarray) -> bool:
    return True if (is_input_2d(values) and is_value(values)) else False


def is_point_2d(points: np.ndarray) -> bool:
    return True if (is_input_2d(points) and is_point(points)) else False
